function [sMap] = ng_batchtrain(sMap, D, varargin)
%NG_BATCHTRAIN Uses batch version to train an unsupervised model 
% using the Neural Gas algorithm.
%
% sM = ngs_batchtrain(sMi,D, [argID, value, ...])
%
% Input and output arguments
% sM       (struct)     Trained model
% sMi      (matrix)     initial position for codebook, copied as initial values
%                       for gradient vectors and 1st dimension will be used as 
%                       initial values for output values
%           (struct)    initial codebook, 
% D         (struct)	training data, data struct
%           (matrix)    training data, size dlen x dim
% [argID,	(string)    Configuration parameters for custom training. 
% value]    (varies)
% 
% Configuration parameters
%  'trainlen' (scalar)  Total number of training epochs
%  'sigma'    (scalar)	Initial neighborhood radius.
%             (vector)	Key values to calculate neighborhood radius. If
%                       length is 2 it will decrease exponentially, if not linearly.
%  'tracking' (scalar)	tracking level 0-3, if negative it will track in
%                       blocks depending on the decimal part.

if ~isstruct(sMap)
    estructura=false;
    Map.codebook=sMap;
    sMap=Map;
    mask=ones(size(sMap.codebook,2),1);
    clear Map;
else
    estructura=true;
    mask=sMap.mask;
    sMap.topol.msize=prod(sMap.topol.msize);

end
    sMap.topol.lattice='none';
    sMap.topol.shape='none';
    sMap.neigh='none';
    
trainstr.type='NG_train';
trainstr.algorithm='unsupervised batch';
if isstruct(D),
    sMap.comp_norm=D.comp_norm;
    sMap.comp_names=D.comp_names;

    trainstr.data_name=D.name;
    D=D.data;
end
%% Map definition values
[munits,dim]= size(sMap.codebook);
dlen = size(D,1);

%% Arguments
error(nargchk(2, Inf, nargin));  % check the number of input arguments
a=1;
while a<=length(varargin),
    if strcmp(varargin(a),'trainlen'), a=a+1; trainlen=varargin{a};end
    if strcmp(varargin(a),'tracking'), a=a+1; tracking=varargin{a};end
    if strcmp(varargin(a),'radius'), a=a+1; sigma=varargin{a};end
    if strcmp(varargin(a),'radius_ini'), a=a+1; sigma=varargin{a};end
    if strcmp(varargin(a),'sigma'), a=a+1; sigma=varargin{a};end
    
    a=a+1;
end

%% Default values if not defined
if ~(exist('trainlen','var')), trainlen=40; end
if ~(exist('tracking','var')), tracking=0; end

%% Neighborhood radius sigma
if ~(exist('sigma','var'))
    % Default value. From half of the codebook units to 0.001
    sigma(1:trainlen)=(munits/2)*(0.002/munits).^((0:trainlen-1)/(trainlen-1));
elseif length(sigma)==1
    % If only a value is given it will be initial radius
    sigma(1:trainlen)=(sigma)*(0.001/sigma).^((0:trainlen-1)/(trainlen-1));
elseif length(sigma)==2
    % If two values are given, they will be initial and final radius
    sigma(1:trainlen)=(sigma(1))*(sigma(2)/sigma(1)).^((0:trainlen-1)/(trainlen-1));
elseif length(sigma)>=trainlen
    % Sigma is keep as it is
else
    % If several values are given, they will be linearly asigned in equal
    % sized periods
    aux=sigma;
    t=round(linspace(1,trainlen,length(aux)));
    for a=1:length(aux)-1, sigma(t(a):t(a+1))=linspace(aux(a),aux(a+1),t(a+1)-t(a)+1); end
end
sigma(sigma==0)=eps; % Avoid div-by-zero error

    trainstr.radius_ini=sigma(1);
    trainstr.radius_fin=sigma(end);
%%
if tracking, qe=zeros(1,trainlen); end
trainstr.trainlen=trainlen;
trainstr.start=datestr(now);
start=clock;
%% For each epoch
for t=1:trainlen
    %% Determine the ranks of the prototipes k(v,wi)
    Hng=zeros(munits,dlen);
    d=zeros(munits,dlen);
    for a=1:dlen
        v=D(a,:);
        d(:,a)=(((v(ones(munits,1),1:dim)-sMap.codebook).^2).*mask(:,ones(munits,1))')*ones(dim,1);
        [~,inds]=sort(d(:,a));
        Hng(inds,a)=0:munits-1;
    end
    Hng=exp(-(Hng)/sigma(t));
    %% Updating prototipes
    numW=Hng*D;
    den=Hng*ones(dlen,1);
    
    nonzero=find(den>0);
    sMap.codebook(nonzero,:)=numW(nonzero,:)./den(nonzero,ones(dim,1));
    %% Tracking
    if tracking,
        qe(t)=real(som_quality(sMap,D));
        trackplot(sMap.codebook ,D,tracking,start,t,qe);
    end
end %for each epoch
    trainstr.end=datestr(now);
if ~estructura, 
    sMap=sMap.codebook; 
else
    sMap.trainhist={sMap.trainhist; trainstr};
end
 sMap.type='Unsupervised Neural Gas Network';
  sMap.name=strcat('Neural Gas Network;',date);
return % function
