function [sMap] = ngs_batchtrain(sMap, D, varargin)
% Copyright (C) 2017. University of Oviedo. All rights reserved.
% Author: Ivn Machn-Gonzlez; Mario Crespo
%
% Please, if you use it, cite the following work:
% Ivn Machn-Gonzlez and Hilario Lpez-Garca, Feedforward Nonlinear Control Using Neural Gas Network, Complexity, vol. 2017, Article ID 3125073, 11 pages, 2017. doi:10.1155/2017/3125073

%NGS_BATCHTRAIN Use batch algorithm to train a supervised model 
% using the NG algorithm.
%
% sM = ngs_batchtrain(sMi,D,[F], [argID, value, ...])
%
% Input and output arguments
% sM       (struct)     Trained model
% sMi      (matrix)     initial position for codebook, copied as initial values
%                       for gradient vectors and 1st dimension will be used as 
%                       initial values for output values
%           (struct)    initial codebook, 
% D         (struct)	training data, data struct
%           (matrix)    training data, size dlen x dim
% F         (matrix)	output values if not included in D (struct) 
% [argID,	(string)    Configuration parameters for custom training. 
% value]    (varies)
% 
% Configuration parameters
%  'trainlen' (scalar)  Total number of training epochs
%  'sigma'    (scalar)	Initial neighborhood radius.
%             (vector)	Key values to calculate neighborhood radius. If
%                       length is 2 it will decrease exponentially, if not linearly.
%  'tracking' (scalar)	tracking level 0-3, if negative it will track in
%                       blocks depending on the decimal part.
%   'values'  (vector)  Initial values for output. Must be (dlen x 1)
%   'vectors' (matrix)  Initial values for gradient components. Must have
%                       same size as the codebook.


if ~isstruct(sMap)
    Map=som_randinit(sMap,'rect','msize',[floor(sqrt(size(sMap,1))) floor(size(sMap,1)/floor(sqrt(size(sMap,1))))]);
    Map.codebook=sMap(1:prod(Map.topol.msize),:);
    Map.vectors=sMap(1:prod(Map.topol.msize),:);
    Map.values=sMap(1:prod(Map.topol.msize),1);
    sMap=Map;
    clear Map;
else
    if ~isfield(sMap,'values'), sMap.values=sMap.codebook(:,1); end
    if ~isfield(sMap,'vectors'), sMap.vectors=sMap.codebook; end
end
trainstr.type='NG_train';
trainstr.algorithm='supervised batch';

if isstruct(D)
    sMap.comp_norm=D.comp_norm;
    sMap.comp_names=D.comp_names;
    sMap.name=D.name;
    trainstr.data_name=D.name;
    if isfield (D,'value_norm'), sMap.value_norm=D.value_norm; end
    if isfield (D,'value_name'), sMap.value_name=D.value_name; else sMap.value_name='Output'; end
    if isfield (D,'values'),
        F=D.values;
        a=1;
    else
        F=varargin{1};
        a=2;
    end
    D=D.data;
else
    F=varargin{1};
    a=2;
end

if ~isnumeric(F), error('Invalid output values or not provided'); end
if size(F,1)~=size(D,1), error('Output values must have same length as input vectors'); end
if size(sMap.codebook,2)~=size(D,2), error('Initial map and data must have the same number of variables (columns)'); end
if any(any(isnan(D))), error ('Data must not contain any NaN'); end
if any(any(isnan(F))), error ('Output values must not contain any NaN'); end

%% Map definition values
[munits,dim]= size(sMap.codebook);
dlen = length (D);

% Default option
sMap.neigh='ranking';

%% Arguments
error(nargchk(2, Inf, nargin));  % check the number of input arguments
while a<=length(varargin),
    %Valued variables
    if strcmp(varargin(a),'trainlen'), a=a+1; trainlen=varargin{a};end
    if strcmp(varargin(a),'tracking'), a=a+1; tracking=varargin{a};end
    if strcmp(varargin(a),'sigma'), a=a+1; sigma=varargin{a}; end
    if strcmp(varargin(a),'radius'), a=a+1; sigma=varargin{a}; end
    if strcmp(varargin(a),'radius_ini'), a=a+1; sigma=varargin{a}; end
    if strcmp(varargin(a),'figure'), a=a+1; Hfig=varargin{a}; end
    
    a=a+1;     
end

%% Default values if not defined
if ~(exist('trainlen','var')), trainlen=50; end
if ~(exist('tracking','var')), tracking=0; end
if ~(exist('fig','var')),
    if abs(tracking)>=2, fig=figure; else fig=NaN;
    end
end
    

%% Neighborhood radius sigma
if ~(exist('sigma','var'))
    % Default value. From half of the codebook units to 0.00001
    sigma(1:trainlen)=(munits/2)*(0.00002/munits).^((0:trainlen-1)/(trainlen-1));
elseif length(sigma)==1
    % If only a value is given it will be initial radius
    sigma(1:trainlen)=(sigma)*(0.00001/sigma).^((0:trainlen-1)/(trainlen-1));
elseif length(sigma)==2
    % If two values are given, they will be initial and final radius
    sigma(1:trainlen)=(sigma(1))*(sigma(2)/sigma(1)).^((0:trainlen-1)/(trainlen-1));
elseif length(sigma)>=trainlen
    % Sigma is keep as it is
else
    % If several values are given, they will be linearly asigned in equal
    % sized periods
    aux=sigma;
    t=round(linspace(1,trainlen,length(aux)));
    for a=1:length(aux)-1, sigma(t(a):t(a+1))=linspace(aux(a),aux(a+1),t(a+1)-t(a)+1); end
end
sigma(sigma==0)=eps; % Avoid div-by-zero error
    trainstr.radius_ini=sigma(1);
    trainstr.radius_fin=sigma(end);

if abs(tracking)>=1, qe=zeros(2,trainlen); end
start=clock;
denA=zeros(munits,1);
trainstr.trainlen=trainlen;
%% For each epoch
trainstr.start=datestr(now);
disp('Training for obtaining prototypes (linearization points)')

for t=1:trainlen

%% Determine the ranks of the prototipes k(v,wi)
    Hng=zeros(munits,dlen);
    d=zeros(munits,dlen);
    for a=1:dlen
        d(:,a)=(D(ones(munits,1)*a,1:dim)-sMap.codebook).^2*ones(dim,1);
        [~,inds]=sort(d(:,a));
        Hng(inds,a)=0:munits-1;
    end
    Hng=exp(-(Hng)/sigma(t));
    
 %% Updating prototypes
    numW=Hng*D;
    numY=Hng*F;
    den=Hng*ones(dlen,1);
 
    nonzero=find(den>eps);
    sMap.codebook(nonzero,:)=numW(nonzero,:)./den(nonzero,ones(dim,1));
    sMap.values(nonzero)=numY(nonzero)./den(nonzero);
  
    %% Tracking
    if abs(tracking)>=1,
         qe(t)=real(som_quality(sMap,D));
      %   trackplot_SNG([sMap.codebook sMap.values],[D F],tracking,start,t,qe,fig);
         trackplot_SNG(tracking,start,t,qe,NaN,trainlen,fig);
     end

end %for each epoch
    
Hng=zeros(munits,dlen); k=Hng;
d=zeros(munits,dlen);
for a=1:dlen
    d(:,a)=(D(ones(munits,1)*a,1:dim)-sMap.codebook).^2*ones(dim,1);
    [~,inds]=sort(d(:,a));
    k(inds,a)=0:(munits-1);
    winners(a)=inds(1);
end

 disp('Training for obtaining gradients')
       
            
for t=1:trainlen
    %% Determine the ranks of the prototipes k(v,wi)
    Hng=exp(-(k)/(sigma(t)));
    %% Updating prototypes
     numA=zeros(munits,dim);
    for a=1:munits
        Dist=D-sMap.codebook(ones(dlen,1)*a,:);
        E=F-sMap.values(ones(dlen,1)*a,:)-((sMap.vectors(ones(dlen,1)*a,:).*Dist)*ones(dim,1));
        numA(a,:)=sum((Hng(ones(dim,1)*a,:)').*E(:,ones(dim,1)).*Dist);
        denA(a)=sum((Hng(a,:)').*(Dist.^2*ones(dim,1)));
    end
        DistWinners=D-sMap.codebook(winners,:);
        ER=F-sMap.values(winners')-((sMap.vectors(winners,:).*DistWinners)*ones(dim,1));
        estimation_error(trainlen+t)=mean(ER.^2);
        nonzero=find(denA>eps);
    sMap.vectors(nonzero,:)=sMap.vectors(nonzero,:)+(numA(nonzero,:)./denA(nonzero,ones(dim,1)));

    %% Tracking
     if abs(tracking)>=1,
       %   [qe(1,trainlen+t),qe(2,trainlen+t)]=som_quality(sMap,D);
        %  qe(:,trainlen+t)=real(qe(:,trainlen+t));
          qe(trainlen:trainlen+t)=som_quality(sMap,D);
         %trackplot_SNG([sMap.codebook sMap.values],[D F],tracking,start,t+trainlen,qe);
         trackplot_SNG(tracking,start,t+trainlen,qe,estimation_error,trainlen);
     end

end %for each epoch

    trainstr.end=datestr(now);
    sMap.trainhist={sMap.trainhist; trainstr};
    sMap.type='Supervised Neural Gas Network';
return % function
